"use strict";
/*--------------------------------------------------------*\
|                                                          |
|                          hprose                          |
|                                                          |
| Official WebSite: https://hprose.com                     |
|                                                          |
| index.ts                                                 |
|                                                          |
| @hprose/rpc-plugin-loadbalance for TypeScript.           |
|                                                          |
| LastModified: Jan 23, 2019                               |
| Author: Ma Bingyao <andot@hprose.com>                    |
|                                                          |
\*________________________________________________________*/
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
function parseWeightedURIList(urilist) {
    var uris = [];
    var weights = [];
    if (urilist) {
        for (var uri in urilist) {
            if (!urilist.hasOwnProperty || urilist.hasOwnProperty(uri)) {
                uris.push(uri);
                var weight = urilist[uri] | 0;
                if (weight <= 0) {
                    throw new Error('Wrong weight value.');
                }
                weights.push(weight);
            }
        }
    }
    return [uris, weights];
}
function randomLoadBalanceHandler(request, context, next) {
    var uris = context.client.uris;
    var n = uris.length;
    context.uri = uris[Math.floor(Math.random() * n)];
    return next(request, context);
}
function getRandomLoadBalanceHandler(urilist) {
    var _this = this;
    var _a = parseWeightedURIList(urilist), uris = _a[0], weights = _a[1];
    if (weights.length === 0) {
        // Random Load Balance
        return randomLoadBalanceHandler;
    }
    // Weighted Random Load Balance
    var effectiveWeights = weights.slice();
    return function (request, context, next) { return __awaiter(_this, void 0, void 0, function () {
        var totalWeight, n, index, currentWeight, i, response, e_1;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    totalWeight = effectiveWeights.reduce(function (x, y) { return x + y; });
                    n = uris.length;
                    index = n - 1;
                    if (totalWeight > 0) {
                        currentWeight = Math.floor(Math.random() * totalWeight);
                        for (i = 0; i < n; ++i) {
                            currentWeight -= effectiveWeights[i];
                            if (currentWeight < 0) {
                                index = i;
                                break;
                            }
                        }
                    }
                    else {
                        index = Math.floor(Math.random() * n);
                    }
                    context.uri = uris[index];
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, next(request, context)];
                case 2:
                    response = _a.sent();
                    if (effectiveWeights[index] < weights[index]) {
                        effectiveWeights[index]++;
                    }
                    return [2 /*return*/, response];
                case 3:
                    e_1 = _a.sent();
                    if (effectiveWeights[index] > 0) {
                        effectiveWeights[index]--;
                    }
                    throw e_1;
                case 4: return [2 /*return*/];
            }
        });
    }); };
}
exports.getRandomLoadBalanceHandler = getRandomLoadBalanceHandler;
function gcd(x, y) {
    var _a, _b;
    if (x < y) {
        _a = [y, x], x = _a[0], y = _a[1];
    }
    while (y !== 0) {
        _b = [y, x % y], x = _b[0], y = _b[1];
    }
    return x;
}
function roundRobinLoadBalanceHandler() {
    //  Round Robin Load Balance
    var index = -1;
    return function (request, context, next) {
        var uris = context.client.uris;
        var n = uris.length;
        if (n > 1) {
            index = (index + 1) % n;
            context.uri = uris[index];
        }
        return next(request, context);
    };
}
function getRoundRobinLoadBalanceHandler(urilist) {
    var _a = parseWeightedURIList(urilist), uris = _a[0], weights = _a[1];
    if (weights.length === 0) {
        //  Round Robin Load Balance
        return roundRobinLoadBalanceHandler();
    }
    // Weighted Round Robin Load Balance
    var maxWeight = Math.max.apply(Math, weights);
    var gcdWeight = weights.reduce(gcd);
    var n = weights.length;
    var index = -1;
    var currentWeight = 0;
    return function (request, context, next) {
        while (true) {
            index = (index + 1) % n;
            if (index === 0) {
                currentWeight -= gcdWeight;
                if (currentWeight <= 0) {
                    currentWeight = maxWeight;
                }
            }
            if (weights[index] >= currentWeight) {
                context.uri = uris[index];
                return next(request, context);
            }
        }
    };
}
exports.getRoundRobinLoadBalanceHandler = getRoundRobinLoadBalanceHandler;
function getNginxRoundRobinLoadBalanceHandler(urilist) {
    var _this = this;
    var _a = parseWeightedURIList(urilist), uris = _a[0], weights = _a[1];
    if (weights.length === 0) {
        //  Round Robin Load Balance
        return roundRobinLoadBalanceHandler();
    }
    // Nginx Weighted Round Robin Load Balance
    var effectiveWeights = weights.slice();
    var n = weights.length;
    var currentWeights = new Array(n).fill(0);
    return function (request, context, next) { return __awaiter(_this, void 0, void 0, function () {
        var _a, totalWeight, index, currentWeight, response, e_2;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    totalWeight = effectiveWeights.reduce(function (x, y) { return x + y; });
                    if (totalWeight > 0) {
                        currentWeight = void 0;
                        _a = effectiveWeights.reduce(function (previous, current, index) {
                            var currentWeight = (currentWeights[index] += current);
                            return (previous[0] < currentWeight) ? [currentWeight, index] : previous;
                        }, [-Infinity, -1]), currentWeight = _a[0], index = _a[1];
                        currentWeights[index] = currentWeight - totalWeight;
                    }
                    else {
                        index = Math.floor(Math.random() * n);
                    }
                    context.uri = uris[index];
                    _b.label = 1;
                case 1:
                    _b.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, next(request, context)];
                case 2:
                    response = _b.sent();
                    if (effectiveWeights[index] < weights[index]) {
                        effectiveWeights[index]++;
                    }
                    return [2 /*return*/, response];
                case 3:
                    e_2 = _b.sent();
                    if (effectiveWeights[index] > 0) {
                        effectiveWeights[index]--;
                    }
                    throw e_2;
                case 4: return [2 /*return*/];
            }
        });
    }); };
}
exports.getNginxRoundRobinLoadBalanceHandler = getNginxRoundRobinLoadBalanceHandler;
function getLeastActiveLoadBalanceHandler(urilist) {
    var _this = this;
    var _a = parseWeightedURIList(urilist), uris = _a[0], weights = _a[1];
    var actives = [];
    if (weights.length === 0) {
        // Least Active Load Balance
        return function (request, context, next) { return __awaiter(_this, void 0, void 0, function () {
            var uris, leastActive, leastActiveIndexes, index, n, response, e_3;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        uris = context.client.uris;
                        if (actives.length === 0) {
                            actives.length = uris.length;
                            actives.fill(0);
                        }
                        leastActive = Math.min.apply(Math, actives);
                        leastActiveIndexes = actives.reduce(function (indexes, active, index) {
                            if (active === leastActive) {
                                indexes.push(index);
                            }
                            return indexes;
                        }, []);
                        index = leastActiveIndexes[0];
                        n = leastActiveIndexes.length;
                        if (n > 1) {
                            index = leastActiveIndexes[Math.floor(Math.random() * n)];
                        }
                        context.uri = uris[index];
                        actives[index]++;
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 3, , 4]);
                        return [4 /*yield*/, next(request, context)];
                    case 2:
                        response = _a.sent();
                        actives[index]--;
                        return [2 /*return*/, response];
                    case 3:
                        e_3 = _a.sent();
                        actives[index]--;
                        throw e_3;
                    case 4: return [2 /*return*/];
                }
            });
        }); };
    }
    // Weighted Least Active Load Balance
    var effectiveWeights = weights.slice();
    actives.length = weights.length;
    actives.fill(0);
    return function (request, context, next) { return __awaiter(_this, void 0, void 0, function () {
        var leastActive, leastActiveIndexes, index, n, totalWeight, currentWeight, i, response, e_4;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    leastActive = Math.min.apply(Math, actives);
                    leastActiveIndexes = actives.reduce(function (indexes, active, index) {
                        if (active === leastActive) {
                            indexes.push(index);
                        }
                        return indexes;
                    }, []);
                    index = leastActiveIndexes[0];
                    n = leastActiveIndexes.length;
                    if (n > 1) {
                        totalWeight = effectiveWeights.reduce(function (x, y, i) { return leastActiveIndexes.indexOf(i) > -1 ? x + y : x; }, 0);
                        if (totalWeight > 0) {
                            currentWeight = Math.floor(Math.random() * totalWeight);
                            for (i = 0; i < n; ++i) {
                                currentWeight -= effectiveWeights[leastActiveIndexes[i]];
                                if (currentWeight < 0) {
                                    index = leastActiveIndexes[i];
                                    break;
                                }
                            }
                        }
                        else {
                            index = leastActiveIndexes[Math.floor(Math.random() * n)];
                        }
                    }
                    context.uri = uris[index];
                    actives[index]++;
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, next(request, context)];
                case 2:
                    response = _a.sent();
                    actives[index]--;
                    if (effectiveWeights[index] < weights[index]) {
                        effectiveWeights[index]++;
                    }
                    return [2 /*return*/, response];
                case 3:
                    e_4 = _a.sent();
                    actives[index]--;
                    if (effectiveWeights[index] > 0) {
                        effectiveWeights[index]--;
                    }
                    throw e_4;
                case 4: return [2 /*return*/];
            }
        });
    }); };
}
exports.getLeastActiveLoadBalanceHandler = getLeastActiveLoadBalanceHandler;
//# sourceMappingURL=index.js.map